"""Libraries related to providing a device level interface for event related media."""

from __future__ import annotations

import datetime
import itertools
import logging
from abc import ABC
from collections import OrderedDict
from collections.abc import Iterable
from typing import Any, Awaitable, Callable, Dict

from .camera_traits import (
    CameraClipPreviewTrait,
    CameraEventImageTrait,
    EventImageContentType,
    EventImageGenerator,
    EventImageType,
)
from .event import (
    CameraMotionEvent,
    CameraPersonEvent,
    CameraSoundEvent,
    DoorbellChimeEvent,
    EventMessage,
    EventToken,
    ImageEventBase,
    session_event_image_type,
)
from .exceptions import GoogleNestException

_LOGGER = logging.getLogger(__name__)

DEFAULT_CACHE_SIZE = 2

# Events are collapsed in the order shown here
VISIBLE_EVENTS = [
    DoorbellChimeEvent.NAME,
    CameraPersonEvent.NAME,
    CameraMotionEvent.NAME,
    CameraSoundEvent.NAME,
]


class CachePolicy:
    """Policy for how many local objects to cache in memory."""

    def __init__(self, event_cache_size: int = DEFAULT_CACHE_SIZE, fetch: bool = False):
        self._event_cache_size = event_cache_size
        self._fetch = fetch
        self._store: EventMediaStore = InMemoryEventMediaStore()

    @property
    def event_cache_size(self) -> int:
        """Number of events to keep in memory per device."""
        return self._event_cache_size

    @event_cache_size.setter
    def event_cache_size(self, value: int) -> None:
        """Set the number of events to keep in memory per device."""
        self._event_cache_size = value

    @property
    def fetch(self) -> bool:
        """Return true if event media should be pre-fetched."""
        return self._fetch

    @fetch.setter
    def fetch(self, value: bool) -> None:
        """Update the value for whether event media should be pre-fetched."""
        self._fetch = value

    @property
    def store(self) -> EventMediaStore:
        """Return the EventMediaStore object."""
        return self._store

    @store.setter
    def store(self, value: EventMediaStore) -> None:
        """Update the EventMediaStore."""
        self._store = value


class Media:
    """Represents media related to an event."""

    def __init__(
        self, contents: bytes, event_image_type: EventImageContentType
    ) -> None:
        """Initialize Media."""
        self._contents = contents
        self._event_image_type = event_image_type

    @property
    def contents(self) -> bytes:
        """Media content."""
        return self._contents

    @property
    def event_image_type(self) -> EventImageContentType:
        """Content event image type of the media."""
        return self._event_image_type

    @property
    def content_type(self) -> str:
        """Content type of the media."""
        return self._event_image_type.content_type


class ImageSession(ABC):
    """An object that holds events that happened within a time range."""

    def __init__(
        self,
        event_session_id: str,
        event_id: str,
        event_timestamp: datetime.datetime,
        event_type: str,
    ) -> None:
        self._event_session_id = event_session_id
        self._event_id = event_id
        self._event_timestamp = event_timestamp
        self._event_type = event_type

    @property
    def event_token(self) -> str:
        token = EventToken(self._event_session_id, self._event_id)
        return token.encode()

    @property
    def timestamp(self) -> datetime.datetime:
        """Return timestamp that the event ocurred."""
        return self._event_timestamp

    @property
    def event_type(self) -> str:
        return self._event_type


class ClipPreviewSession(ABC):
    """An object that holds events that happened within a time range."""

    def __init__(
        self,
        event_session_id: str,
        event_timestamp: datetime.datetime,
        event_types: list[str],
    ) -> None:
        self._event_session_id = event_session_id
        self._event_timestamp = event_timestamp
        self._event_types = event_types

    @property
    def event_token(self) -> str:
        token = EventToken(self._event_session_id)
        return token.encode()

    @property
    def timestamp(self) -> datetime.datetime:
        """Return timestamp that the event ocurred."""
        return self._event_timestamp

    @property
    def event_types(self) -> list[str]:
        return self._event_types


class EventMedia:
    """Represents an event and its associated media."""

    def __init__(
        self,
        event_session_id: str,
        event_id: str,
        event_type: str,
        event_timestamp: datetime.datetime,
        media: Media,
    ) -> None:
        self._event_session_id = event_session_id
        self._event_id = event_id
        self._event_type = event_type
        self._event_timestamp = event_timestamp
        self._media = media

    @property
    def event_session_id(self) -> str:
        """Return the event session id."""
        return self._event_session_id

    @property
    def event_type(self) -> str:
        """Return the event type."""
        return self._event_type

    @property
    def event_timestamp(self) -> datetime.datetime:
        """Return timestamp that the event ocurred."""
        return self._event_timestamp

    @property
    def media(self) -> Media:
        return self._media


class EventMediaStore(ABC):
    """Interface for external storage."""

    async def async_load(self) -> dict | None:
        """Load data."""

    async def async_save(self, data: dict) -> None:
        """Save data."""

    def get_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename to use for the device and event."""

    def get_image_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename for image media."""
        return self.get_media_key(device_id, event)

    def get_clip_preview_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the filename for clip preview media."""
        return self.get_media_key(device_id, event)

    async def async_load_media(self, media_key: str) -> bytes | None:
        """Load media content."""

    async def async_save_media(self, media_key: str, content: bytes) -> None:
        """Write media content."""

    async def async_remove_media(self, media_key: str) -> None:
        """Remove media content."""


class InMemoryEventMediaStore(EventMediaStore):
    """An in memory implementation of EventMediaStore."""

    def __init__(self) -> None:
        self._data: dict | None = None
        self._media: dict[str, bytes] = {}

    async def async_load(self) -> dict | None:
        """Load data."""
        return self._data

    async def async_save(self, data: dict) -> None:
        """Save data."""
        self._data = data

    def get_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        suffix = "jpg" if event.event_image_type == EventImageType.IMAGE else "mp4"
        return f"{device_id}_{event.timestamp}_{event.event_session_id}.{suffix}"

    def get_image_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        return f"{device_id}_{event.timestamp}_{event.event_session_id}_{event.event_id}.jpg"

    def get_clip_preview_media_key(self, device_id: str, event: ImageEventBase) -> str:
        """Return the media key to use for the device and event."""
        return f"{device_id}_{event.timestamp}_{event.event_session_id}.mp4"

    async def async_load_media(self, media_key: str) -> bytes | None:
        """Load media content."""
        return self._media.get(media_key)

    async def async_save_media(self, media_key: str, content: bytes) -> None:
        """Remove media content."""
        self._media[media_key] = content

    async def async_remove_media(self, media_key: str) -> None:
        """Remove media content."""
        if media_key in self._media:
            del self._media[media_key]


class EventMediaModelItem:
    """Structure used to persist the event in EventMediaStore."""

    def __init__(
        self,
        event_session_id: str,
        events: dict[str, ImageEventBase],
        media_key: str | None,
        event_media_keys: dict[str, str],
    ) -> None:
        """Initialize EventMediaModelItem."""
        self._event_session_id = event_session_id
        self._events = events
        self._media_key = media_key
        self._event_media_keys = event_media_keys if event_media_keys else {}

    @staticmethod
    def from_dict(data: dict[str, Any]) -> EventMediaModelItem:
        """Read from serialized dictionary."""
        events: dict[str, ImageEventBase] = {}
        input_events = data.get("events", {})
        for (event_type, event_data) in input_events.items():
            if not (event := ImageEventBase.from_dict(event_data)):
                continue
            events[event_type] = event
        # Link events to other events in the session
        event_image_type = session_event_image_type(events.values())
        for event in events.values():
            event.session_events = list(events.values())
            event.event_image_type = event_image_type
        return EventMediaModelItem(
            data["event_session_id"],
            events,
            data.get("media_key"),
            data.get("event_media_keys", {}),
        )

    @property
    def event_session_id(self) -> str:
        return self._event_session_id

    @property
    def visible_event(self) -> ImageEventBase | None:
        """Get the primary visible event for this item."""
        for event_type in VISIBLE_EVENTS:
            if event := self._events.get(event_type):
                return event
        return None

    @property
    def events(self) -> dict[str, ImageEventBase]:
        """Return all associated events with this record."""
        return self._events

    @property
    def media_key(self) -> str | None:
        return self._media_key

    @media_key.setter
    def media_key(self, value: str) -> None:
        """Update the media_key."""
        self._media_key = value

    @property
    def event_media_keys(self) -> dict[str, str]:
        return self._event_media_keys

    def media_key_for_token(self, token: EventToken) -> str | None:
        """Return media key for the specified event token."""
        if token.event_id:
            if token.event_id in self._event_media_keys:
                return self._event_media_keys[token.event_id]
            # Fallback to legacy single event per session
        return self._media_key

    @property
    def any_media_key(self) -> str | None:
        """Return any media item for compatibility with legacy APIs."""
        if self._media_key:
            return self._media_key
        if self._event_media_keys.values():
            return next(iter(self._event_media_keys.values()))
        return None

    def get_media(self, content: bytes) -> Media:
        assert self.visible_event
        return Media(content, self.visible_event.event_image_type)

    def get_event_media(self, content: bytes) -> EventMedia:
        assert self.visible_event
        return EventMedia(
            self.visible_event.event_session_id,
            self.visible_event.event_id,
            self.visible_event.event_type,
            self.visible_event.timestamp,
            self.get_media(content),
        )

    def as_dict(self) -> dict[str, Any]:
        """Return a EventMediaModelItem as a serializable dict."""
        result: dict[str, Any] = {
            "event_session_id": self._event_session_id,
            "events": dict((k, v.as_dict()) for k, v in self._events.items()),
            "event_media_keys": self._event_media_keys,
        }
        if self._media_key:
            result["media_key"] = self._media_key
        return result

    def __repr__(self) -> str:
        return (
            "<EventMediaModelItem events="
            + str(self._events)
            + " media_key="
            + str(self._media_key)
            + ">"
        )


class EventMediaManager:
    """Responsible for handling recent events and fetching associated media."""

    def __init__(
        self,
        device_id: str,
        traits: Dict[str, Any],
        event_trait_map: Dict[str, EventImageGenerator],
        support_fetch: bool,
    ) -> None:
        """Initialize DeviceEventMediaManager."""
        self._device_id = device_id
        self._traits = traits
        self._event_trait_map = event_trait_map
        self._cache_policy = CachePolicy()
        self._callback: Callable[[EventMessage], Awaitable[None]] | None = None
        self._support_fetch = support_fetch

    @property
    def cache_policy(self) -> CachePolicy:
        """Return the current CachePolicy."""
        return self._cache_policy

    @cache_policy.setter
    def cache_policy(self, value: CachePolicy) -> None:
        """Update the CachePolicy."""
        self._cache_policy = value

    async def _async_load(self) -> OrderedDict[str, EventMediaModelItem]:
        """Load the device specific data from the store."""
        store_data = await self._cache_policy.store.async_load()
        event_data: OrderedDict[str, EventMediaModelItem] = OrderedDict()
        if store_data:
            device_data = store_data.get(self._device_id, [])
            for item_data in device_data:
                item = EventMediaModelItem.from_dict(item_data)
                event_data[item.event_session_id] = item
        return event_data

    async def _async_update(self, event_data: dict[str, EventMediaModelItem]) -> None:
        """Save the device specific model to the store."""
        # Event order is preserved so popping from the oldest entry works
        device_data: list[dict[str, Any]] = []
        for item in event_data.values():
            device_data.append(item.as_dict())

        # Read data from the store and update information for this device
        store_data = await self._cache_policy.store.async_load()
        if not store_data:
            store_data = {}
        store_data[self._device_id] = device_data
        await self._cache_policy._store.async_save(store_data)

    async def get_media(
        self, event_session_id: str, width: int | None = None, height: int | None = None
    ) -> EventMedia | None:
        """Get media for the specified event.

        Note that the height and width hints are best effort and may not be
        honored (e.g. if image is already cached).
        """
        event_data = await self._async_load()
        if not (item := event_data.get(event_session_id)):
            return None
        event = item.visible_event
        if not event:
            _LOGGER.debug(
                "Skipping fetch; No visible event; event_session_id=%s",
                event_session_id,
            )
            return None
        media_key = item.any_media_key
        if media_key:
            contents = await self._cache_policy._store.async_load_media(media_key)
            if contents:
                return item.get_event_media(contents)

        if event.is_expired:
            _LOGGER.debug(
                "Skipping fetch; Event expired; event_session_id=%s",
                event_session_id,
            )
            return None
        _LOGGER.debug("Fetching media for event_session_id=%s", event_session_id)
        if not (generator := self._event_trait_map.get(event.event_type)):
            return None
        event_image = await generator.generate_event_image(event)
        if not event_image:
            return None
        contents = await event_image.contents(width=width, height=height)
        media_key = self._cache_policy._store.get_media_key(self._device_id, event)
        await self._cache_policy.store.async_save_media(media_key, contents)
        item.media_key = media_key
        await self._async_update(event_data)
        return item.get_event_media(contents)

    async def _fetch_media(self, item: EventMediaModelItem) -> None:
        """Fetch media from the server in response to a pubsub event."""
        store = self._cache_policy.store
        if CameraClipPreviewTrait.NAME in self._traits:
            if (
                item.media_key
                or not item.visible_event
                or item.visible_event.is_expired
            ):
                return
            clip_preview_trait: CameraClipPreviewTrait = self._traits[
                CameraClipPreviewTrait.NAME
            ]
            event_image = await clip_preview_trait.generate_event_image(
                item.visible_event
            )
            if event_image:
                content = await event_image.contents()
                # Caller will persist the media key assignment
                media_key = store.get_clip_preview_media_key(
                    self._device_id, item.visible_event
                )
                item.media_key = media_key
                _LOGGER.debug("Saving media %s (%s)", media_key, item.event_session_id)
                await store.async_save_media(media_key, content)
                return

        if CameraEventImageTrait.NAME not in self._traits:
            return

        event_image_trait: CameraEventImageTrait = self._traits[
            CameraEventImageTrait.NAME
        ]
        for event in item.events.values():
            if event.event_id in item.event_media_keys or event.is_expired:
                continue
            event_image = await event_image_trait.generate_image(event.event_id)
            content = await event_image.contents()

            # Caller will persist the media key assignment
            media_key = store.get_image_media_key(self._device_id, event)
            item.event_media_keys[event.event_id] = media_key
            _LOGGER.debug("Saving media %s (%s)", media_key, item.event_session_id)
            await store.async_save_media(media_key, content)

    async def get_media_from_token(self, event_token: str) -> Media | None:
        """Get media based on the event token."""
        token = EventToken.decode(event_token)
        event_data = await self._async_load()
        if not (item := event_data.get(token.event_session_id)):
            _LOGGER.debug(
                "No event information found for event id: %s", token.event_session_id
            )
            return None
        media_key = item.media_key_for_token(token)
        if not media_key:
            _LOGGER.debug("No persisted media for event id %s", token)
            return None
        contents = await self._cache_policy._store.async_load_media(media_key)
        if not contents:
            _LOGGER.debug(
                "Unable to load persisted media for event id: (%s, %s, %s)",
                token.event_session_id,
                token.event_id,
                item.media_key,
            )
            return None
        assert item.visible_event
        return Media(contents, item.visible_event.event_image_type)

    async def async_events(self) -> Iterable[ImageEventBase]:
        """Return revent events."""
        result = await self._visible_items()

        def _get_event(x: EventMediaModelItem) -> ImageEventBase:
            assert x.visible_event
            return x.visible_event

        event_result = list(map(_get_event, result))
        event_result.sort(key=lambda x: x.timestamp, reverse=True)
        return event_result

    async def async_image_sessions(self) -> Iterable[ImageSession]:
        """Return revent events."""

        result = await self._visible_items()

        def _get_events(x: EventMediaModelItem) -> list[ImageEventBase]:
            return list(x.events.values())

        events_list = list(map(_get_events, result))
        events: Iterable[ImageEventBase] = itertools.chain(*events_list)

        def _get_session(x: ImageEventBase) -> ImageSession:
            return ImageSession(
                x.event_session_id, x.event_id, x.timestamp, x.event_type
            )

        event_result = list(map(_get_session, events))
        event_result.sort(key=lambda x: x.timestamp, reverse=True)
        return event_result

    async def async_clip_preview_sessions(self) -> Iterable[ClipPreviewSession]:
        """Return revent events."""

        result = await self._visible_items()

        def _event_visible(event_type: str) -> bool:
            return event_type in VISIBLE_EVENTS

        def _get_event_session(x: EventMediaModelItem) -> ClipPreviewSession:
            assert x.visible_event
            events = list(x.events.values())
            events.sort(key=lambda x: x.timestamp)
            return ClipPreviewSession(
                x.event_session_id,
                next(iter(events)).timestamp,
                list(filter(_event_visible, [y.event_type for y in events])),
            )

        event_result = list(map(_get_event_session, result))
        event_result.sort(key=lambda x: x.timestamp, reverse=True)
        return event_result

    async def _visible_items(self) -> list[EventMediaModelItem]:
        """Return items in the modle that are visible events for serving."""

        def _filter(x: EventMediaModelItem) -> bool:
            """Return events already fetched or that could be fetched."""
            return x.visible_event is not None and (
                x.media_key is not None or not x.visible_event.is_expired
            )

        event_data = await self._async_load()
        return list(filter(_filter, event_data.values()))

    def set_update_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> None:
        """Register a callback invoked when new messages are received."""
        self._callback = target

    async def async_handle_events(self, event_message: EventMessage) -> None:
        """Handle the EventMessage."""
        event_sessions: dict[
            str, dict[str, ImageEventBase]
        ] | None = event_message.event_sessions
        if not event_sessions:
            return
        _LOGGER.debug("Event Update %s", event_sessions.keys())

        # Notify traits to cache most recent event
        pairs = list(event_sessions.items())
        for event_session_id, event_dict in pairs:
            supported = False
            for event_name, event in event_dict.items():
                if not (trait := self._event_trait_map.get(event_name)):
                    _LOGGER.debug("Unsupported event trait: %s", event_name)
                    continue
                supported = True
                trait.handle_event(event)

            # Skip any entirely unsupported events
            if not supported:
                del event_sessions[event_session_id]

        # Update interal event media representation. Events are only published
        # to downstream subscribers the first time they are seen to avoid
        # firing on updated event threads multiple times.
        suppress_keys: set[str] = set({})
        valid_events = 0
        for event_session_id, event_dict in event_sessions.items():

            # Track all related events together with the same session
            event_data = await self._async_load()

            if model_item := event_data.get(event_session_id):
                # Update the existing event session with new/updated events. Only
                # new events are published.
                suppress_keys |= set(model_item.events.keys())
                valid_events += len(
                    set(event_dict.keys()) - set(model_item.events.keys())
                )
                model_item.events.update(event_dict)
            else:
                # A new event session
                valid_events += len(event_dict.keys())
                model_item = EventMediaModelItem(
                    event_session_id, event_dict, media_key=None, event_media_keys={}
                )
                event_data[event_session_id] = model_item

                if len(event_data) > self._cache_policy.event_cache_size:
                    (key, old_item) = event_data.popitem(last=False)
                    if old_item.media_key:
                        _LOGGER.debug(
                            "Expiring media %s (%s)",
                            old_item.media_key,
                            old_item.event_session_id,
                        )
                        await self._cache_policy.store.async_remove_media(
                            old_item.media_key
                        )
                    for old_media_key in old_item.event_media_keys.values():
                        _LOGGER.debug(
                            "Expiring event media %s (%s)",
                            old_media_key,
                            old_item.event_session_id,
                        )
                        await self._cache_policy.store.async_remove_media(old_media_key)

            await self._async_update(event_data)

            if self._support_fetch and self._cache_policy.fetch:
                try:
                    await self._fetch_media(model_item)
                except GoogleNestException as err:
                    _LOGGER.warning(
                        "Failure when pre-fetching event '%s': %s",
                        event.event_session_id,
                        str(err),
                    )
                # Update any new media keys
                await self._async_update(event_data)

        if not self._callback:
            return
        # Notify any listeners about the arrival of a new event
        if suppress_keys:
            event_message = event_message.omit_events(suppress_keys)
        if valid_events > 0:
            await self._callback(event_message)

    @property
    def active_event_trait(self) -> EventImageGenerator | None:
        """Return trait with the most recently received active event."""
        trait_to_return: EventImageGenerator | None = None
        for trait in self._event_trait_map.values():
            if not trait.active_event:
                continue
            if trait_to_return is None:
                trait_to_return = trait
            else:
                event = trait.last_event
                if not event or not trait_to_return.last_event:
                    continue
                if event.expires_at > trait_to_return.last_event.expires_at:
                    trait_to_return = trait
        return trait_to_return

    async def get_active_event_media(
        self, width: int | None = None, height: int | None = None
    ) -> EventMedia | None:
        """Return a snapshot image for a very recent event if any."""
        if not (trait := self.active_event_trait):
            return None
        event: ImageEventBase | None = trait.last_event
        if not event:
            return None
        return await self.get_media(event.event_session_id, width, height)



"""Events from pubsub subscriber."""

from __future__ import annotations

import base64
import binascii
import datetime
import hashlib
import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Mapping, Optional

from .auth import AbstractAuth
from .exceptions import DecodeException
from .registry import Registry
from .traits import BuildTraits, Command
from .typing import cast_assert, cast_optional

EVENT_ID = "eventId"
EVENT_SESSION_ID = "eventSessionId"
TIMESTAMP = "timestamp"
RESOURCE_UPDATE = "resourceUpdate"
NAME = "name"
TRAITS = "traits"
EVENTS = "events"
RELATION_UPDATE = "relationUpdate"
TYPE = "type"
SUBJECT = "subject"
OBJECT = "object"
PREVIEW_URL = "previewUrl"

# Event images expire 30 seconds after the event is published
EVENT_IMAGE_EXPIRE_SECS = 30

# Camera clip previews don't list an expiration in the API. Lets say 15 minutes
# as an arbitrary number for now.
CAMERA_CLIP_PREVIEW_EXPIRE_SECS = 15 * 60

EVENT_MAP = Registry()

_LOGGER = logging.getLogger(__name__)


class EventProcessingError(Exception):
    """Raised when there was an error handling an event."""


class EventImageContentType:
    """Event image content type."""

    def __init__(self, content_type: str) -> None:
        """Initialize EventImageContentType."""
        self._content_type = content_type

    def __str__(self) -> str:
        """Return a string representation of the event image type."""
        return self._content_type

    def __repr__(self) -> str:
        return "<EventImageContentType %s>" % self._content_type

    @property
    def content_type(self) -> str:
        """Return the content type of the image."""
        return self._content_type


class EventImageType:
    IMAGE = EventImageContentType("image/jpeg")
    CLIP_PREVIEW = EventImageContentType("video/mp4")
    IMAGE_PREVIEW = EventImageContentType("image/gif")

    @staticmethod
    def from_string(content_type: str) -> EventImageContentType:
        """Parse an EventImageType from a string representation."""
        if content_type == EventImageType.CLIP_PREVIEW.content_type:
            return EventImageType.CLIP_PREVIEW
        elif content_type == EventImageType.IMAGE.content_type:
            return EventImageType.IMAGE
        elif content_type == EventImageType.IMAGE_PREVIEW.content_type:
            return EventImageType.IMAGE_PREVIEW
        else:
            return EventImageContentType(content_type)


@dataclass
class EventToken:
    """Identifier for a unique event."""

    event_session_id: str
    event_id: str | None = None

    def encode(self) -> str:
        """Encode the event token as a serialized string."""
        data = [self.event_session_id, self.event_id]
        b = json.dumps(data).encode("utf-8")
        return base64.b64encode(b).decode("utf-8")

    @staticmethod
    def decode(content: str) -> EventToken:
        """Decode an event token into a class."""
        try:
            s = base64.b64decode(content).decode("utf-8")
        except binascii.Error as err:
            raise DecodeException from err
        data = json.loads(s)
        if not isinstance(data, list) or len(data) != 2:
            raise DecodeException("Unexpected data type: %s", data)
        return EventToken(data[0], data[1])

    def __repr__(self) -> str:
        if not self.event_id:
            return "<EventToken event_session_id" + self.event_session_id + ">"
        return (
            "<EventToken event_session_id"
            + self.event_session_id
            + " event_id="
            + self.event_id
            + ">"
        )


class ImageEventBase(ABC):
    """Base class for all image related event types."""

    event_image_type: EventImageContentType

    def __init__(self, data: Mapping[str, Any], timestamp: datetime.datetime) -> None:
        """Initialize EventBase."""
        self._data = data
        self._timestamp = timestamp
        self._session_events: list[ImageEventBase] = []

    @property
    def event_token(self) -> str:
        """An identifier of this session / event combination."""
        token = EventToken(self.event_session_id, self.event_id)
        return token.encode()

    @property
    @abstractmethod
    def event_type(self) -> str:
        """The type of event."""

    @property
    def event_id(self) -> str:
        """ID associated with the event.

        Can be used with CameraEventImageTrait to download the imaage.
        """
        return cast_assert(str, self._data[EVENT_ID])

    @property
    def event_session_id(self) -> str:
        """ID used to associate separate messages with a single event."""
        return cast_assert(str, self._data[EVENT_SESSION_ID])

    @property
    def timestamp(self) -> datetime.datetime:
        """Timestap when the event occurred."""
        return self._timestamp

    @property
    def expires_at(self) -> datetime.datetime:
        """Timestamp when the message expires."""
        return self._timestamp + datetime.timedelta(seconds=EVENT_IMAGE_EXPIRE_SECS)

    @property
    def is_expired(self) -> bool:
        """Return true if the event expiration has passed."""
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        return self.expires_at < now

    @property
    def session_events(self) -> list[ImageEventBase]:
        return self._session_events

    @session_events.setter
    def session_events(self, value: list[ImageEventBase]) -> None:
        self._session_events = value

    def as_dict(self) -> dict[str, Any]:
        """Return as a dict form that can be serialized."""
        return {
            "event_type": self.event_type,
            "event_data": self._data,
            "timestamp": self._timestamp.isoformat(),
            "event_image_type": str(self.event_image_type),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> ImageEventBase | None:
        """Parse from a serialized dictionary."""
        event_type = data["event_type"]
        event_data = data["event_data"]
        timestamp = datetime.datetime.fromisoformat(data["timestamp"])
        event = _BuildEvent(event_type, event_data, timestamp)
        if event and "event_image_type" in data:
            event.event_image_type = EventImageType.from_string(
                data["event_image_type"]
            )
        return event

    def __repr__(self) -> str:
        return (
            "<ImageEventBase "
            + str(self.as_dict())
            + " sessions="
            + str(len(self._session_events))
            + ">"
        )


@EVENT_MAP.register()
class CameraMotionEvent(ImageEventBase):
    """Motion has been detected by the camera."""

    NAME = "sdm.devices.events.CameraMotion.Motion"
    event_type = NAME
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraPersonEvent(ImageEventBase):
    """A person has been detected by the camera."""

    NAME = "sdm.devices.events.CameraPerson.Person"
    event_type = NAME
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraSoundEvent(ImageEventBase):
    """Sound has been detected by the camera."""

    NAME = "sdm.devices.events.CameraSound.Sound"
    event_type = NAME
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class DoorbellChimeEvent(ImageEventBase):
    """The doorbell has been pressed."""

    NAME = "sdm.devices.events.DoorbellChime.Chime"
    event_type = NAME
    event_image_type = EventImageType.IMAGE


@EVENT_MAP.register()
class CameraClipPreviewEvent(ImageEventBase):
    """A video clip is available for preview, without extra download."""

    NAME = "sdm.devices.events.CameraClipPreview.ClipPreview"
    event_type = NAME
    event_image_type = EventImageType.CLIP_PREVIEW

    @property
    def event_id(self) -> str:
        """Use a URL hash as the event id.

        Since clip preview events already have a url associated with them,
        we don't have an event id for downloading the image.
        """
        return hashlib.blake2b(self.preview_url.encode()).hexdigest()

    @property
    def expires_at(self) -> datetime.datetime:
        """Event ids do not expire."""
        return self._timestamp + datetime.timedelta(
            seconds=CAMERA_CLIP_PREVIEW_EXPIRE_SECS
        )

    @property
    def preview_url(self) -> str:
        """A url 10 second frame video file in mp4 format."""
        return cast_assert(str, self._data[PREVIEW_URL])


class EventTrait(ABC):
    """Parent class for traits related to handling events."""

    def __init__(self) -> None:
        """Initialize an EventTrait."""
        self._last_event: Optional[ImageEventBase] = None

    @property
    def last_event(self) -> Optional[ImageEventBase]:
        """Last received event."""
        return self._last_event

    @property
    def active_event(self) -> Optional[ImageEventBase]:
        """Any current active events."""
        if not self._last_event:
            return None
        if self._last_event.is_expired:
            return None
        return self._last_event

    def handle_event(self, event: ImageEventBase) -> None:
        """Recieve an event message."""
        self._last_event = event


class RelationUpdate:
    """Represents a relational update for a resource."""

    def __init__(self, raw_data: Mapping[str, Any]):
        """Initialize the RelationUpdate."""
        self._raw_data = raw_data

    @property
    def type(self) -> str:
        """Type of relation event 'CREATED', 'UPDATED', 'DELETED'."""
        return cast_assert(str, self._raw_data[TYPE])

    @property
    def subject(self) -> str:
        """Resource that the object is now in relation with."""
        return cast_assert(str, self._raw_data.get(SUBJECT))

    @property
    def object(self) -> str:
        """Resource that triggered the event."""
        return cast_assert(str, self._raw_data[OBJECT])


def _BuildEvents(
    events: Mapping[str, Any],
    timestamp: datetime.datetime,
) -> Dict[str, ImageEventBase]:
    """Build a trait map out of a response dict."""
    result = {}
    for (event_type, event_data) in events.items():
        image_event = _BuildEvent(event_type, event_data, timestamp)
        if not image_event:
            continue
        result[event_type] = image_event
    return result


def _BuildEvent(
    event_type: str, event_data: Mapping[str, Any], timestamp: datetime.datetime
) -> ImageEventBase | None:
    if event_type not in EVENT_MAP:
        return None
    cls = EVENT_MAP[event_type]
    return cls(event_data, timestamp)  # type: ignore


def session_event_image_type(events: Iterable[ImageEventBase]) -> EventImageContentType:
    """Determine the event type to use based on the events in the session."""
    for event in events:
        if event.event_image_type != EventImageType.IMAGE:
            return event.event_image_type
    return EventImageType.IMAGE


class EventMessage:
    """Event for a change in trait value or device action."""

    def __init__(self, raw_data: Mapping[str, Any], auth: AbstractAuth):
        """Initialize an EventMessage."""
        _LOGGER.debug("EventMessage raw_data=%s", raw_data)
        self._raw_data = raw_data
        self._auth = auth

    @property
    def event_id(self) -> Optional[str]:
        """Event identifier."""
        return self._raw_data.get(EVENT_ID)

    @property
    def timestamp(self) -> datetime.datetime:
        """Time when the event was published."""
        event_timestamp = self._raw_data[TIMESTAMP]
        return datetime.datetime.fromisoformat(event_timestamp.replace("Z", "+00:00"))

    @property
    def resource_update_name(self) -> Optional[str]:
        """Return the id of the device that was updated."""
        if RESOURCE_UPDATE not in self._raw_data:
            return None
        return cast_optional(str, self._raw_data[RESOURCE_UPDATE].get(NAME))

    @property
    def resource_update_events(self) -> Optional[Dict[str, ImageEventBase]]:
        """Return the set of events that happened."""
        if RESOURCE_UPDATE not in self._raw_data:
            return None
        events = self._raw_data[RESOURCE_UPDATE].get(EVENTS, {})
        assert isinstance(events, dict)
        return _BuildEvents(events, self.timestamp)

    @property
    def event_sessions(self) -> Optional[dict[str, dict[str, ImageEventBase]]]:
        events = self.resource_update_events
        if not events:
            return None
        event_sessions: dict[str, dict[str, ImageEventBase]] = {}
        for (event_name, event) in events.items():
            d = event_sessions.get(event.event_session_id, {})
            d[event_name] = event
            event_sessions[event.event_session_id] = d
        # Build associations between all events
        for event_session_id, event_dict in event_sessions.items():
            event_image_type = session_event_image_type(events.values())
            for event_type, event in event_dict.items():
                event.session_events = list(event_dict.values())
                event.event_image_type = event_image_type
        return event_sessions

    @property
    def resource_update_traits(self) -> Optional[dict]:
        """Return the set of traits that were updated."""
        if not self.resource_update_name:
            return None
        cmd = Command(self.resource_update_name, self._auth)
        events = self._raw_data[RESOURCE_UPDATE].get(TRAITS, {})
        return BuildTraits(events, cmd)

    @property
    def relation_update(self) -> Optional[RelationUpdate]:
        """Represent a relational update for a resource."""
        if RELATION_UPDATE not in self._raw_data:
            return None
        return RelationUpdate(self._raw_data[RELATION_UPDATE])

    @property
    def raw_data(self) -> Dict[str, Any]:
        """Return raw data for the event."""
        return dict(self._raw_data)

    def omit_events(self, omit_event_keys: Iterable[str]) -> EventMessage:
        """Create a new EventMessage minus some existing events by key."""
        raw_data = dict(self._raw_data)
        events = raw_data[RESOURCE_UPDATE].get(EVENTS, {})
        for key in omit_event_keys:
            if key not in events:
                continue
            del events[key]
        raw_data[RESOURCE_UPDATE][EVENTS] = events
        return EventMessage(raw_data, self._auth)



"""Subscriber for the Smart Device Management event based API."""
import asyncio
import concurrent.futures
import json
import logging
import re
from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Optional

from aiohttp.client_exceptions import ClientError
from google.api_core.exceptions import GoogleAPIError, NotFound, Unauthenticated
from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.cloud import pubsub_v1
from google.oauth2.credentials import Credentials
from google.protobuf.duration_pb2 import Duration

from .auth import AbstractAuth
from .device_manager import DeviceManager
from .event import EventMessage
from .event_media import CachePolicy
from .exceptions import AuthException, ConfigurationException, SubscriberException
from .google_nest_api import GoogleNestAPI

_LOGGER = logging.getLogger(__name__)

# Used to catch invalid subscriber id
EXPECTED_SUBSCRIBER_REGEXP = re.compile("projects/.*/subscriptions/.*")

# Used to catch a topic misconfiguration
EXPECTED_TOPIC_REGEXP = re.compile("projects/sdm-[a-z]+/topics/.*")

WATCHDOG_CHECK_INTERVAL_SECONDS = 10

# Restart the subscriber after some delay, with exponential backoff
WATCHDOG_RESTART_DELAY_MIN_SECONDS = 10
WATCHDOG_RESTART_DELAY_MAX_SECONDS = 300
# Reset watchdog backoff
WATCHDOG_RESET_THRESHOLD_SECONDS = 60

DEFAULT_MESSAGE_RETENTION_SECONDS = 15 * 60  # 15 minutes

# Note: Users of non-prod instances will have to manually configure a topic
TOPIC_FORMAT = "projects/sdm-prod/topics/enterprise-{project_id}"

OAUTH2_AUTHORIZE_FORMAT = (
    "https://nestservices.google.com/partnerconnections/{project_id}/auth"
)
OAUTH2_TOKEN = "https://www.googleapis.com/oauth2/v4/token"
SDM_SCOPES = [
    "https://www.googleapis.com/auth/sdm.service",
    "https://www.googleapis.com/auth/pubsub",
]
API_URL = "https://smartdevicemanagement.googleapis.com/v1"


def _validate_subscription_name(subscription_name: str) -> None:
    """Validates that a subcription name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_SUBSCRIBER_REGEXP.match(subscription_name):
        raise ConfigurationException(
            "Subscription misconfigured. Expected subscriber_id to "
            f"match '{EXPECTED_SUBSCRIBER_REGEXP.pattern}' but was "
            f"'{subscription_name}'"
        )


def _validate_topic_name(topic_name: str) -> None:
    """Validates that a topic name is correct.

    Raises ConfigurationException on failure.
    """
    if not EXPECTED_TOPIC_REGEXP.match(topic_name):
        raise ConfigurationException(
            "Subscription misconfigured. Expected topic name to "
            f"match '{EXPECTED_TOPIC_REGEXP.pattern}' but was "
            f"'{topic_name}'."
        )


class AbstractSubscriberFactory(ABC):
    """Abstract class for creating a subscriber, to facilitate testing."""

    @abstractmethod
    async def async_create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""

    @abstractmethod
    async def async_delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Deletes a subscription."""

    @abstractmethod
    async def async_new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
        async_callback: Callable[
            [pubsub_v1.subscriber.message.Message], Awaitable[None]
        ],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Create a new event subscriber."""


class DefaultSubscriberFactory(AbstractSubscriberFactory):
    """Default implementation that creates Google Pubsub subscriber."""

    async def async_create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        _validate_subscription_name(subscription_name)
        _validate_topic_name(topic_name)
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            await loop.run_in_executor(
                executor,
                self._create_subscription,
                creds,
                subscription_name,
                topic_name,
            )

    def _create_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        topic_name: str,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        creds = self._refresh_creds(creds)
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)

        subscription = None
        try:
            subscription = subscriber.get_subscription(subscription=subscription_name)
        except NotFound:
            _LOGGER.debug("Existing subscription not found; Creating")
        if subscription:
            if subscription.topic != topic_name:
                raise ConfigurationException(
                    "Subscription misconfigured. Expected topic name to "
                    f"match '{topic_name}' but was "
                    f"'{subscription.topic}'. Please delete in cloud "
                    "console and it will be re-created."
                )
            # Valid subscription already exists; No-op
            return

        message_retention_duration = Duration()
        message_retention_duration.FromSeconds(DEFAULT_MESSAGE_RETENTION_SECONDS)
        subscription_request = {
            "name": subscription_name,
            "topic": topic_name,
            "message_retention_duration": message_retention_duration,
        }
        _LOGGER.debug(f"Creating subscription: {subscription_request}")
        subscriber.create_subscription(request=subscription_request)

    async def async_delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """Creates a subscription name if it does not already exist."""
        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            await loop.run_in_executor(
                executor,
                self._delete_subscription,
                creds,
                subscription_name,
            )

    def _delete_subscription(
        self,
        creds: Credentials,
        subscription_name: str,
    ) -> None:
        """Deletes a subscription."""
        creds = self._refresh_creds(creds)
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)
        _LOGGER.debug(f"Deleting subscription '{subscription_name}'")
        subscriber.delete_subscription(subscription=subscription_name)

    async def async_new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        loop: asyncio.AbstractEventLoop,
        async_callback: Callable[
            [pubsub_v1.subscriber.message.Message], Awaitable[None]
        ],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Create a new subscriber with a blocking to async bridge."""

        def callback_wrapper(message: pubsub_v1.subscriber.message.Message) -> None:
            if loop.is_closed():
                return
            assert async_callback
            future: concurrent.futures.Future = asyncio.run_coroutine_threadsafe(
                async_callback(message), loop
            )
            future.result()

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            return await loop.run_in_executor(
                executor,
                self._new_subscriber,
                creds,
                subscription_name,
                callback_wrapper,
            )

    def _refresh_creds(self, creds: Credentials) -> Credentials:
        """Refresh credentials."""
        try:
            creds.refresh(Request())
        except RefreshError as err:
            raise AuthException(f"Access token failure: {err}") from err
        return creds

    def _new_subscriber(
        self,
        creds: Credentials,
        subscription_name: str,
        callback_wrapper: Callable[[pubsub_v1.subscriber.message.Message], None],
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Issue a command to verify subscriber creds are correct."""
        creds = self._refresh_creds(creds)
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)
        subscription = subscriber.get_subscription(subscription=subscription_name)
        if subscription.topic:
            _validate_topic_name(subscription.topic)
            _LOGGER.debug(
                "Subscriber '%s' configured on topic '%s'",
                subscription_name,
                subscription.topic,
            )
        return subscriber.subscribe(subscription_name, callback_wrapper)


class GoogleNestSubscriber:
    """Subscribe to events from the Google Nest feed."""

    def __init__(
        self,
        auth: AbstractAuth,
        project_id: str,
        subscriber_id: str,
        subscriber_factory: AbstractSubscriberFactory = DefaultSubscriberFactory(),
        loop: Optional[asyncio.AbstractEventLoop] = None,
        watchdog_check_interval_seconds: float = WATCHDOG_CHECK_INTERVAL_SECONDS,
        watchdog_restart_delay_min_seconds: float = WATCHDOG_RESTART_DELAY_MIN_SECONDS,
    ):
        """Initialize the subscriber for the specified topic."""
        self._auth = auth
        self._subscriber_id = subscriber_id
        self._project_id = project_id
        self._api = GoogleNestAPI(auth, project_id)
        self._loop = loop or asyncio.get_event_loop()
        self._device_manager_task: Optional[asyncio.Task[DeviceManager]] = None
        self._subscriber_factory = subscriber_factory
        self._subscriber_future: Optional[
            pubsub_v1.subscriber.futures.StreamingPullFuture
        ] = None
        self._callback: Optional[Callable[[EventMessage], Awaitable[None]]] = None
        self._healthy = True
        self._watchdog_check_interval_seconds = watchdog_check_interval_seconds
        self._watchdog_restart_delay_min_seconds = watchdog_restart_delay_min_seconds
        self._watchdog_restart_delay_seconds = watchdog_restart_delay_min_seconds
        self._watchdog_task: Optional[asyncio.Task] = None
        self._cache_policy = CachePolicy()

    @property
    def subscriber_id(self) -> str:
        """Return the configured subscriber_id."""
        return self._subscriber_id

    @property
    def project_id(self) -> str:
        """Return the configured SDM project_id."""
        return self._project_id

    def set_update_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> None:
        """Register a callback invoked when new messages are received."""
        self._callback = target
        if self._device_manager_task and self._device_manager_task.done():
            self._device_manager_task.result().set_update_callback(target)

    async def create_subscription(self) -> None:
        """Create the subscription if it does not already exist."""
        _validate_subscription_name(self._subscriber_id)
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err
        try:
            await self._subscriber_factory.async_create_subscription(
                creds,
                self._subscriber_id,
                TOPIC_FORMAT.format(project_id=self._project_id),
                self._loop,
            )
        except NotFound as err:
            raise ConfigurationException(
                f"Failed to create subscriber '{self._subscriber_id}' "
                + "(cloud project id incorrect?)"
            ) from err
        except Unauthenticated as err:
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}': {err}"
            ) from err

    async def delete_subscription(self) -> None:
        """Delete the subscription."""
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err

        try:
            await self._subscriber_factory.async_delete_subscription(
                creds, self._subscriber_id, self._loop
            )
        except NotFound:
            # No-op if subscription was already deleted
            return
        except Unauthenticated as err:
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            raise SubscriberException(
                f"Failed to delete subscription '{self._subscriber_id}': {err}"
            ) from err

    async def start_async(self) -> None:
        """Start the subscriber."""
        _validate_subscription_name(self._subscriber_id)
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err

        try:
            self._subscriber_future = (
                await self._subscriber_factory.async_new_subscriber(
                    creds, self._subscriber_id, self._loop, self._async_message_callback
                )
            )
        except NotFound as err:
            raise ConfigurationException(
                f"Failed to create subscriber '{self._subscriber_id}' id was not found"
            ) from err
        except Unauthenticated as err:
            raise AuthException("Failed to authenticate subscriber: {err}") from err
        except GoogleAPIError as err:
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}': {err}"
            ) from err

        if not self._healthy:
            _LOGGER.debug("Subscriber reconnected")
            self._healthy = True
            self._watchdog_restart_delay_seconds = (
                self._watchdog_restart_delay_min_seconds
            )
        if self._watchdog_check_interval_seconds > 0:
            self._watchdog_task = asyncio.create_task(self._watchdog())
        if self._subscriber_future:
            self._subscriber_future.add_done_callback(self._done_callback)

    async def _watchdog(self) -> None:
        """Background task that watches the subscriber and restarts it."""
        _LOGGER.debug("Starting background watchdog thread")
        while True:
            if self._subscriber_future and self._subscriber_future.done():
                _LOGGER.debug(
                    "Watchdog: subscriber shut down; restarting in %s seconds",
                    self._watchdog_restart_delay_seconds,
                )
                await asyncio.sleep(self._watchdog_restart_delay_seconds)
                self._watchdog_restart_delay_seconds *= 2
                self._watchdog_restart_delay_seconds = max(
                    self._watchdog_restart_delay_seconds,
                    WATCHDOG_RESTART_DELAY_MAX_SECONDS,
                )
                await self.start_async()
            await asyncio.sleep(self._watchdog_check_interval_seconds)

    def stop_async(self) -> None:
        """Tell the subscriber to start shutting down."""
        if self._device_manager_task:
            self._device_manager_task.cancel()
        if self._watchdog_task:
            self._watchdog_task.cancel()
        if self._subscriber_future:
            self._subscriber_future.cancel()

    @property
    def cache_policy(self) -> CachePolicy:
        """Return cache policy shared by device EventMediaManager objects."""
        return self._cache_policy

    async def async_get_device_manager(self) -> DeviceManager:
        """Return the DeviceManger with the current state of devices."""
        if not self._device_manager_task:
            self._device_manager_task = asyncio.create_task(
                self._async_create_device_manager()
            )
        assert self._device_manager_task
        return await self._device_manager_task

    async def _async_create_device_manager(self) -> DeviceManager:
        """Create a DeviceManager, populated with initial state."""
        device_manager = DeviceManager(self._cache_policy)
        structures = await self._api.async_get_structures()
        for structure in structures:
            device_manager.add_structure(structure)
        # Subscriber starts after a device fetch
        devices = await self._api.async_get_devices()
        for device in devices:
            device_manager.add_device(device)
        if self._callback:
            device_manager.set_update_callback(self._callback)
        return device_manager

    def _done_callback(
        self, future: pubsub_v1.subscriber.futures.StreamingPullFuture
    ) -> None:
        """Teardown subscriber."""
        if future.exception():
            ex = future.exception()
            if self._healthy:
                self._healthy = False
            _LOGGER.debug("Subscriber disconnected, will restart: %s: %s", type(ex), ex)

    async def _async_message_callback(
        self, message: pubsub_v1.subscriber.message.Message
    ) -> None:
        """Handle a received message."""
        payload = json.loads(bytes.decode(message.data))
        event = EventMessage(payload, self._auth)
        # Only accept device events once the Device Manager has been loaded.
        # We are ok with missing messages on startup since the device manager
        # will do a live read .
        if self._device_manager_task and self._device_manager_task.done():
            await self._device_manager_task.result().async_handle_event(event)
        message.ack()



"""Library for traits about devices."""

import datetime
from typing import Any, Dict, Mapping, Optional, cast

import aiohttp

from .traits import TRAIT_MAP, Command
from .typing import cast_assert, cast_optional

STATUS = "status"
TIMER_MODE = "timerMode"
TIMER_TIMEOUT = "timerTimeout"
CUSTOM_NAME = "customName"
AMBIENT_HUMIDITY_PERCENT = "ambientHumidityPercent"
AMBIENT_TEMPERATURE_CELSIUS = "ambientTemperatureCelsius"


@TRAIT_MAP.register()
class ConnectivityTrait:
    """This trait belongs to any device that has connectivity information."""

    NAME = "sdm.devices.traits.Connectivity"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize ConnectivityTrait."""
        self._data = data

    @property
    def status(self) -> str:
        """Device connectivity status.

        Return:
          "OFFLINE", "ONLINE"
        """
        return cast_assert(str, self._data[STATUS])


@TRAIT_MAP.register()
class FanTrait:
    """This trait belongs to any device that can control the fan."""

    NAME = "sdm.devices.traits.Fan"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize FanTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def timer_mode(self) -> Optional[str]:
        """Timer mode for the fan.

        Return:
          "ON", "OFF"
        """
        return cast_optional(str, self._data.get(TIMER_MODE))

    @property
    def timer_timeout(self) -> Optional[datetime.datetime]:
        """Timestamp at which timer mode will turn to OFF."""
        if TIMER_TIMEOUT not in self._data:
            return None
        timestamp = self._data[TIMER_TIMEOUT]
        return datetime.datetime.fromisoformat(timestamp.replace("Z", "+00:00"))

    async def set_timer(
        self, timer_mode: str, duration: Optional[int] = None
    ) -> aiohttp.ClientResponse:
        """Change the fan timer."""
        data: Dict[str, Any] = {
            "command": "sdm.devices.commands.Fan.SetTimer",
            "params": {
                "timerMode": timer_mode,
            },
        }
        if duration:
            data["params"]["duration"] = f"{duration}s"
        return await self._cmd.execute(data)


@TRAIT_MAP.register()
class InfoTrait:
    """This trait belongs to any device for device-related information."""

    NAME = "sdm.devices.traits.Info"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize InfoTrait."""
        self._data = data

    @property
    def custom_name(self) -> Optional[str]:
        """Name of the device."""
        return cast_optional(str, self._data.get(CUSTOM_NAME))


@TRAIT_MAP.register()
class HumidityTrait:
    """This trait belongs to any device that has a sensor to measure humidity."""

    NAME = "sdm.devices.traits.Humidity"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize HumidityTrait."""
        self._data = data

    @property
    def ambient_humidity_percent(self) -> float:
        """Percent humidity, measured at the device."""
        return cast(float, self._data[AMBIENT_HUMIDITY_PERCENT])


@TRAIT_MAP.register()
class TemperatureTrait:
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME = "sdm.devices.traits.Temperature"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize TemperatureTrait."""
        self._data = data

    @property
    def ambient_temperature_celsius(self) -> float:
        """Percent humidity, measured at the device."""
        return cast(float, self._data[AMBIENT_TEMPERATURE_CELSIUS])



"""Library to access the Smart Device Management API."""
from typing import List, Optional

from .auth import AbstractAuth
from .device import Device
from .structure import Structure

STRUCTURES = "structures"
DEVICES = "devices"
NAME = "name"


class GoogleNestAPI:
    """Class to communicate with the Google Nest SDM API."""

    def __init__(self, auth: AbstractAuth, project_id: str):
        """Initialize the API and store the auth so we can make requests."""
        self._auth = auth
        self._project_id = project_id

    @property
    def _structures_url(self) -> str:
        return f"enterprises/{self._project_id}/structures"

    async def async_get_structures(self) -> List[Structure]:
        """Return the structures."""
        response_data = await self._auth.get_json(self._structures_url)
        if STRUCTURES not in response_data:
            return []
        structures = response_data[STRUCTURES]
        return [
            Structure.MakeStructure(structure_data) for structure_data in structures
        ]

    async def async_get_structure(self, structure_id: str) -> Optional[Structure]:
        """Return a structure device."""
        data = await self._auth.get_json(f"{self._structures_url}/{structure_id}")
        if NAME not in data:
            return None
        return Structure.MakeStructure(data)

    @property
    def _devices_url(self) -> str:
        return f"enterprises/{self._project_id}/devices"

    async def async_get_devices(self) -> List[Device]:
        """Return the devices."""
        response_data = await self._auth.get_json(self._devices_url)
        if DEVICES not in response_data:
            return []
        devices = response_data[DEVICES]
        return [Device.MakeDevice(device_data, self._auth) for device_data in devices]

    async def async_get_device(self, device_id: str) -> Optional[Device]:
        """Return a specific device."""
        data = await self._auth.get_json(f"{self._devices_url}/{device_id}")
        if NAME not in data:
            return None
        return Device.MakeDevice(data, self._auth)



"""Device Manager keeps track of the current state of all devices."""

from __future__ import annotations

from typing import Awaitable, Callable, Dict, Optional

from .device import Device
from .event import EventMessage, RelationUpdate
from .event_media import CachePolicy
from .structure import InfoTrait, RoomInfoTrait, Structure


class DeviceManager:
    """DeviceManager holds current state of all devices."""

    def __init__(self, cache_policy: CachePolicy | None = None) -> None:
        """Initialize DeviceManager."""
        self._devices: Dict[str, Device] = {}
        self._structures: Dict[str, Structure] = {}
        self._cache_policy = cache_policy if cache_policy else CachePolicy()
        self._callback: Callable[[EventMessage], Awaitable[None]] | None = None

    @property
    def devices(self) -> Dict[str, Device]:
        """Return current state of devices."""
        return self._devices

    @property
    def structures(self) -> Dict[str, Structure]:
        """Return current state of structures."""
        return self._structures

    def add_device(self, device: Device) -> None:
        """Track the specified device."""
        assert device.name
        self._devices[device.name] = device
        # Share a single cache policy across all devices
        device.event_media_manager.cache_policy = self._cache_policy
        if self._callback:
            device.event_media_manager.set_update_callback(self._callback)

    def add_structure(self, structure: Structure) -> None:
        """Track the specified device."""
        assert structure.name
        self._structures[structure.name] = structure

    @property
    def cache_policy(self) -> CachePolicy:
        """Return cache policy shared by device EventMediaManager objects."""
        return self._cache_policy

    def set_update_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> None:
        """Register a callback invoked when new messages are received."""
        self._callback = target
        for device in self._devices.values():
            device.event_media_manager.set_update_callback(target)

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Handle a new message received."""
        if event_message.relation_update:
            self._handle_device_relation(event_message.relation_update)
            if self._callback:
                await self._callback(event_message)
            return

        if event_message.resource_update_name:
            device_id = event_message.resource_update_name
            if device_id in self._devices:
                device = self._devices[device_id]
                await device.async_handle_event(event_message)

    def _structure_name(self, relation_subject: str) -> Optional[str]:
        if relation_subject in self._structures:
            structure = self._structures[relation_subject]
            for trait_name in [InfoTrait.NAME, RoomInfoTrait.NAME]:
                if trait_name in structure.traits:
                    return structure.traits[trait_name].custom_name
        return "Unknown"

    def _handle_device_relation(self, relation: RelationUpdate) -> None:
        if relation.object not in self._devices:
            return

        device = self._devices[relation.object]
        if relation.type == "DELETED":
            # Delete device from room/structure
            if relation.subject in device.parent_relations:
                del device.parent_relations[relation.subject]

        if relation.type == "UPDATED" or relation.type == "CREATED":
            # Device moved to a room
            assert relation.subject
            device.parent_relations[relation.subject] = self._structure_name(
                relation.subject
            )



"""Authentication library, implemented by users of the API."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, List, Mapping, Optional

import aiohttp
from aiohttp.client_exceptions import ClientError, ClientResponseError
from google.auth.credentials import Credentials
from google.oauth2.credentials import Credentials as OAuthCredentials

from .exceptions import ApiException, AuthException

_LOGGER = logging.getLogger(__name__)

HTTP_UNAUTHORIZED = 401
AUTHORIZATION_HEADER = "Authorization"
ERROR = "error"
STATUS = "status"
MESSAGE = "message"


class AbstractAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: aiohttp.ClientSession, host: str):
        """Initialize the auth."""
        self._websession = websession
        self._host = host

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def async_get_creds(self) -> Credentials:
        """Return creds for subscriber API."""
        token = await self.async_get_access_token()
        return OAuthCredentials(token=token)

    async def request(
        self, method: str, url: str, **kwargs: Optional[Mapping[str, Any]]
    ) -> aiohttp.ClientResponse:
        """Make a request."""
        headers = kwargs.get("headers")

        if headers is None:
            headers = {}
        else:
            headers = dict(headers)
            del kwargs["headers"]
        if AUTHORIZATION_HEADER not in headers:
            try:
                access_token = await self.async_get_access_token()
            except ClientError as err:
                raise AuthException(f"Access token failure: {err}") from err
            headers[AUTHORIZATION_HEADER] = f"Bearer {access_token}"
        if not (url.startswith("http://") or url.startswith("https://")):
            url = f"{self._host}/{url}"
        _LOGGER.debug("request[%s]=%s", method, url)
        if method == "post" and "json" in kwargs:
            _LOGGER.debug("request[post json]=%s", kwargs["json"])
        return await self._websession.request(method, url, **kwargs, headers=headers)

    async def get(
        self, url: str, **kwargs: Mapping[str, Any]
    ) -> aiohttp.ClientResponse:
        """Make a get request."""
        try:
            resp = await self.request("get", url, **kwargs)
        except ClientError as err:
            raise ApiException(f"Error connecting to API: {err}") from err
        return await AbstractAuth._raise_for_status(resp)

    async def get_json(self, url: str, **kwargs: Mapping[str, Any]) -> dict[str, Any]:
        """Make a get request and return json response."""
        resp = await self.get(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        if not isinstance(result, dict):
            raise ApiException("Server return malformed response: %s" % result)
        _LOGGER.debug("response=%s", result)
        return result

    async def post(
        self, url: str, **kwargs: Mapping[str, Any]
    ) -> aiohttp.ClientResponse:
        """Make a post request."""
        try:
            resp = await self.request("post", url, **kwargs)
        except ClientError as err:
            raise ApiException(f"Error connecting to API: {err}") from err
        return await AbstractAuth._raise_for_status(resp)

    async def post_json(self, url: str, **kwargs: Mapping[str, Any]) -> dict[str, Any]:
        """Make a post request and return a json response."""
        resp = await self.post(url, **kwargs)
        try:
            result = await resp.json()
        except ClientError as err:
            raise ApiException("Server returned malformed response") from err
        if not isinstance(result, dict):
            raise ApiException("Server returned malformed response: %s" % result)
        _LOGGER.debug("response=%s", result)
        return result

    @staticmethod
    async def _raise_for_status(resp: aiohttp.ClientResponse) -> aiohttp.ClientResponse:
        """Raise exceptions on failure methods."""
        detail = await AbstractAuth._error_detail(resp)
        try:
            resp.raise_for_status()
        except ClientResponseError as err:
            if err.status == HTTP_UNAUTHORIZED:
                raise AuthException(f"Unable to authenticate with API: {err}") from err
            detail.append(err.message)
            raise ApiException(": ".join(detail)) from err
        except ClientError as err:
            raise ApiException(f"Error from API: {err}") from err
        return resp

    @staticmethod
    async def _error_detail(resp: aiohttp.ClientResponse) -> List[str]:
        """Resturns an error message string from the APi response."""
        if resp.status < 400:
            return []
        try:
            result = await resp.json()
            error = result.get(ERROR, {})
        except ClientError:
            return []
        message = ["Error from API", f"{resp.status}"]
        if STATUS in error:
            message.append(f"{error[STATUS]}")
        if MESSAGE in error:
            message.append(error[MESSAGE])
        return message



"""Traits belonging to doorbell devices."""

from __future__ import annotations

import logging
from typing import Any, Mapping, Optional

from .camera_traits import EventImage, EventImageCreator, EventImageGenerator
from .event import DoorbellChimeEvent, ImageEventBase
from .traits import TRAIT_MAP, Command

_LOGGER = logging.getLogger(__name__)


@TRAIT_MAP.register()
class DoorbellChimeTrait(EventImageGenerator):
    """For any device that supports a doorbell chime and related press events."""

    NAME = "sdm.devices.traits.DoorbellChime"
    EVENT_NAME = DoorbellChimeEvent.NAME
    event_type = EVENT_NAME
    event_image_creator: EventImageCreator | None = None

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize DoorbellChime."""
        super().__init__()
        self._data = data

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        _LOGGER.debug("Generating image for event")
        if not isinstance(event, DoorbellChimeEvent):
            return None
        assert event.event_id
        if not self.event_image_creator:
            raise ValueError("Camera does not have trait to fetch snapshots")
        return await self.event_image_creator.generate_event_image(event)



"""Traits for thermostats."""

from __future__ import annotations

from abc import ABC
from typing import Any, Final, List, Mapping, cast

import aiohttp

from .traits import TRAIT_MAP, Command
from .typing import cast_assert

STATUS: Final = "status"
AVAILABLE_MODES: Final = "availableModes"
MODE: Final = "mode"
HEAT_CELSIUS: Final = "heatCelsius"
COOL_CELSIUS: Final = "coolCelsius"


class ThermostatHeatCoolTrait(ABC):
    """Parent class for traits related to temperature set points."""

    @property
    def heat_celsius(self) -> float:
        """Lowest temperature where Eco mode begins heating."""

    @property
    def cool_celsius(self) -> float:
        """Highest cooling temperature where Eco mode begins cooling."""


@TRAIT_MAP.register()
class ThermostatEcoTrait(ThermostatHeatCoolTrait):
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME = "sdm.devices.traits.ThermostatEco"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize ThermostatEcoTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def available_modes(self) -> List[str]:
        """List of supported Eco modes."""
        return cast(List[str], self._data[AVAILABLE_MODES])

    @property
    def mode(self) -> str:
        """Eco mode of the thermostat."""
        return cast_assert(str, self._data[MODE])

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatEco.SetMode",
            "params": {"mode": mode},
        }
        return await self._cmd.execute(data)

    @property
    def heat_celsius(self) -> float:
        """Lowest temperature where Eco mode begins heating."""
        return cast(float, self._data[HEAT_CELSIUS])

    @property
    def cool_celsius(self) -> float:
        """Highest cooling temperature where Eco mode begins cooling."""
        return cast(float, self._data[COOL_CELSIUS])


@TRAIT_MAP.register()
class ThermostatHvacTrait:
    """This trait belongs to devices that can report HVAC details."""

    NAME = "sdm.devices.traits.ThermostatHvac"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize ThermostatHvacTrait."""
        self._data = data

    @property
    def status(self) -> List[str]:
        """HVAC status of the thermostat."""
        return cast(List[str], self._data[STATUS])


@TRAIT_MAP.register()
class ThermostatModeTrait:
    """This trait belongs to devices that support different thermostat modes."""

    NAME = "sdm.devices.traits.ThermostatMode"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize ThermostatModeTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def available_modes(self) -> List[str]:
        """List of supported thermostat modes."""
        return cast(List[str], self._data[AVAILABLE_MODES])

    @property
    def mode(self) -> str:
        """Mode of the thermostat."""
        return cast_assert(str, self._data[MODE])

    async def set_mode(self, mode: str) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatMode.SetMode",
            "params": {"mode": mode},
        }
        return await self._cmd.execute(data)


@TRAIT_MAP.register()
class ThermostatTemperatureSetpointTrait(ThermostatHeatCoolTrait):
    """This trait belongs to devices that support setting target temperature."""

    NAME = "sdm.devices.traits.ThermostatTemperatureSetpoint"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize ThermostatTemperatureSetpointTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def heat_celsius(self) -> float:
        """Lowest temperature where Eco mode begins heating."""
        return cast(float, self._data[HEAT_CELSIUS])

    @property
    def cool_celsius(self) -> float:
        """Highest cooling temperature where Eco mode begins cooling."""
        return cast(float, self._data[COOL_CELSIUS])

    async def set_heat(self, heat: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetHeat",
            "params": {"heatCelsius": heat},
        }
        return await self._cmd.execute(data)

    async def set_cool(self, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetCool",
            "params": {"coolCelsius": cool},
        }
        return await self._cmd.execute(data)

    async def set_range(self, heat: float, cool: float) -> aiohttp.ClientResponse:
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetRange",
            "params": {
                "heatCelsius": heat,
                "coolCelsius": cool,
            },
        }
        return await self._cmd.execute(data)



"""Traits for structures / rooms."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Mapping, Optional, cast

from .registry import Registry
from .typing import cast_optional

STRUCTURE_NAME = "name"
STRUCTURE_TRAITS = "traits"
CUSTOM_NAME = "customName"

STRUCTURE_TRAITS_MAP = Registry()


class StructureTrait(ABC):
    """This trait belongs to any structure for structure-related information."""

    @property
    @abstractmethod
    def custom_name(self) -> Optional[str]:
        """Name of the structure."""


@STRUCTURE_TRAITS_MAP.register()
class InfoTrait(StructureTrait):
    """This trait belongs to any structure for structure-related information."""

    NAME = "sdm.structures.traits.Info"

    def __init__(self, data: Mapping[str, Any]):
        """Initialize InfoTrait."""
        self._data = data

    @property
    def custom_name(self) -> Optional[str]:
        """Name of the structure."""
        return cast_optional(str, self._data[CUSTOM_NAME])


@STRUCTURE_TRAITS_MAP.register()
class RoomInfoTrait(StructureTrait):
    """This trait belongs to any structure for room-related information."""

    NAME = "sdm.structures.traits.RoomInfo"

    def __init__(self, data: Mapping[str, Any]) -> None:
        """Initialize RoomInfoTrait."""
        self._data = data

    @property
    def custom_name(self) -> Optional[str]:
        """Name of the room."""
        return cast_optional(str, self._data[CUSTOM_NAME])


def _TraitsDict(
    traits: Mapping[str, Any], trait_map: Mapping[str, Any]
) -> Dict[str, StructureTrait]:
    result: Dict[str, StructureTrait] = {}
    for (trait, trait_data) in traits.items():
        if trait not in trait_map:
            continue
        cls = trait_map[trait]
        result[trait] = cls(trait_data)
    return result


class Structure:
    """Class that represents a structure object in the Google Nest SDM API."""

    def __init__(
        self, raw_data: Mapping[str, Any], traits: Mapping[str, StructureTrait]
    ) -> None:
        """Initialize a structure."""
        self._raw_data = raw_data
        self._traits = traits

    @staticmethod
    def MakeStructure(raw_data: Mapping[str, Any]) -> Structure:
        """Create a structure with the appropriate traits."""
        traits = raw_data.get(STRUCTURE_TRAITS, {})
        traits_dict = _TraitsDict(traits, STRUCTURE_TRAITS_MAP)
        return Structure(raw_data, traits_dict)

    @property
    def name(self) -> Optional[str]:
        """Resource name of the structure e.g. 'enterprises/XYZ/structures/123'."""
        return cast_optional(str, self._raw_data[STRUCTURE_NAME])

    @property
    def traits(self) -> Dict[str, StructureTrait]:
        """Return a trait mixin on None."""
        return dict(self._traits)

    def _traits_data(self, trait: Mapping[str, Any]) -> Dict[str, Any]:
        """Return the raw dictionary for the specified trait."""
        traits_dict = self._raw_data.get(STRUCTURE_TRAITS, {})
        return cast(Dict[str, Any], traits_dict.get(trait, {}))

    @property
    def raw_data(self) -> Dict[str, Any]:
        """Return raw data for the structure."""
        return dict(self._raw_data)



"""Traits belonging to camera devices."""

from __future__ import annotations

import datetime
import logging
import urllib.parse as urlparse
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, List, Mapping, Optional, cast

from .event import (
    CameraClipPreviewEvent,
    CameraMotionEvent,
    CameraPersonEvent,
    CameraSoundEvent,
    EventImageContentType,
    EventImageType,
    EventTrait,
    ImageEventBase,
)
from .traits import TRAIT_MAP, Command
from .typing import cast_assert, cast_optional

_LOGGER = logging.getLogger(__name__)

MAX_IMAGE_RESOLUTION = "maxImageResolution"
MAX_VIDEO_RESOLUTION = "maxVideoResolution"
WIDTH = "width"
HEIGHT = "height"
VIDEO_CODECS = "videoCodecs"
AUDIO_CODECS = "audioCodecs"
SUPPORTED_PROTOCOLS = "supportedProtocols"
STREAM_URLS = "streamUrls"
RESULTS = "results"
RTSP_URL = "rtspUrl"
STREAM_EXTENSION_TOKEN = "streamExtensionToken"
STREAM_TOKEN = "streamToken"
URL = "url"
TOKEN = "token"
EXPIRES_AT = "expiresAt"
ANSWER_SDP = "answerSdp"
MEDIA_SESSION_ID = "mediaSessionId"

EVENT_IMAGE_CLIP_PREVIEW = "clip_preview"


class Resolution:
    """Maximum Resolution of an image or stream."""

    width: Optional[int] = None
    height: Optional[int] = None


@TRAIT_MAP.register()
class CameraImageTrait:
    """This trait belongs to any device that supports taking images."""

    NAME = "sdm.devices.traits.CameraImage"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraImageTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def max_image_resolution(self) -> Resolution:
        """Maximum resolution of the camera image."""
        res = Resolution()
        res.width = self._data[MAX_IMAGE_RESOLUTION][WIDTH]
        res.height = self._data[MAX_IMAGE_RESOLUTION][HEIGHT]
        return res


class Stream(ABC):
    """Base class for streams."""

    def __init__(self, data: Mapping[str, Any]):
        self._data = data

    @property
    def expires_at(self) -> datetime.datetime:
        """Time at which both streamExtensionToken and streamToken expire."""
        expires_at = self._data[EXPIRES_AT]
        return datetime.datetime.fromisoformat(expires_at.replace("Z", "+00:00"))

    @abstractmethod
    async def extend_stream(self) -> Stream:
        """Extend the lifetime of the stream."""

    @abstractmethod
    async def stop_stream(self) -> None:
        """Invalidate the stream."""


class RtspStream(Stream):
    """Provides access an RTSP live stream URL."""

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize RstpStream."""
        super().__init__(data)
        self._cmd = cmd

    @property
    def rtsp_stream_url(self) -> str:
        """RTSP live stream URL."""
        rtsp_stream_url = self._data[STREAM_URLS][RTSP_URL]
        assert isinstance(rtsp_stream_url, str)
        return rtsp_stream_url

    @property
    def stream_token(self) -> str:
        """Token to use to access an RTSP live stream."""
        stream_token = self._data[STREAM_TOKEN]
        assert isinstance(stream_token, str)
        return stream_token

    @property
    def stream_extension_token(self) -> str:
        """Token to use to access an RTSP live stream."""
        return cast_assert(str, self._data[STREAM_EXTENSION_TOKEN])

    async def extend_stream(self) -> Stream | RtspStream:
        """Extend the lifetime of the stream."""
        return await self.extend_rtsp_stream()

    async def extend_rtsp_stream(self) -> RtspStream:
        """Request a new RTSP live stream URL access token."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.ExtendRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        response_data = await self._cmd.execute_json(data)
        results = response_data[RESULTS]
        # Update the stream url with the new token
        stream_token = results[STREAM_TOKEN]
        parsed = urlparse.urlparse(self.rtsp_stream_url)
        parsed = parsed._replace(query=f"auth={stream_token}")
        url = urlparse.urlunparse(parsed)
        results[STREAM_URLS] = {}
        results[STREAM_URLS][RTSP_URL] = url
        return RtspStream(results, self._cmd)

    async def stop_stream(self) -> None:
        """Invalidate the stream."""
        return await self.stop_rtsp_stream()

    async def stop_rtsp_stream(self) -> None:
        """Invalidates a valid RTSP access token and stops the RTSP live stream."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.StopRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        await self._cmd.execute(data)


class WebRtcStream(Stream):
    """Provides access an RTSP live stream URL."""

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize RstpStream."""
        super().__init__(data)
        self._cmd = cmd

    @property
    def answer_sdp(self) -> str:
        """An SDP answer to use with the local device dispalying the stream."""
        answer_sdp = self._data[ANSWER_SDP]
        assert isinstance(answer_sdp, str)
        return answer_sdp

    @property
    def media_session_id(self) -> str:
        """Media Session ID of the live stream."""
        media_session_id = self._data[MEDIA_SESSION_ID]
        assert isinstance(media_session_id, str)
        return media_session_id

    async def extend_stream(self) -> WebRtcStream:
        """Request a new RTSP live stream URL access token."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.ExtendWebRtcStream",
            "params": {MEDIA_SESSION_ID: self.media_session_id},
        }
        response_data = await self._cmd.execute_json(data)
        # Preserve original answerSdp, and merge with response that contains
        # the other fields (expiresAt, and mediaSessionId.
        results = response_data[RESULTS]
        results[ANSWER_SDP] = self.answer_sdp
        return WebRtcStream(results, self._cmd)

    async def stop_stream(self) -> None:
        """Invalidates a valid RTSP access token and stops the RTSP live stream."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.StopWebRtcStream",
            "params": {MEDIA_SESSION_ID: self.media_session_id},
        }
        await self._cmd.execute(data)


class StreamingProtocol(Enum):
    """Streaming protocols supported by the device."""

    RTSP = (1,)
    WEB_RTC = (2,)


@TRAIT_MAP.register()
class CameraLiveStreamTrait:
    """This trait belongs to any device that supports live streaming."""

    NAME = "sdm.devices.traits.CameraLiveStream"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraLiveStreamTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def max_video_resolution(self) -> Resolution:
        """Maximum resolution of the video live stream."""
        res = Resolution()
        res.width = self._data[MAX_VIDEO_RESOLUTION][WIDTH]
        res.height = self._data[MAX_VIDEO_RESOLUTION][HEIGHT]
        return res

    @property
    def video_codecs(self) -> List[str]:
        """Video codecs supported for the live stream."""
        assert isinstance(self._data[VIDEO_CODECS], list)
        return cast(List[str], self._data[VIDEO_CODECS])

    @property
    def audio_codecs(self) -> List[str]:
        """Audio codecs supported for the live stream."""
        return cast(List[str], self._data[AUDIO_CODECS])

    @property
    def supported_protocols(self) -> List[StreamingProtocol]:
        """Streaming protocols supported for the live stream."""
        return [
            StreamingProtocol[x]
            for x in self._data.get(SUPPORTED_PROTOCOLS, ["RTSP"])
            if x in StreamingProtocol.__members__
        ]

    async def generate_rtsp_stream(self) -> RtspStream:
        """Request a token to access an RTSP live stream URL."""
        if StreamingProtocol.RTSP not in self.supported_protocols:
            raise ValueError("Device does not support RTSP stream")
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.GenerateRtspStream",
            "params": {},
        }
        response_data = await self._cmd.execute_json(data)
        results = response_data[RESULTS]
        return RtspStream(results, self._cmd)

    async def generate_web_rtc_stream(self, offer_sdp: str) -> WebRtcStream:
        """Request a token to access a Web RTC live stream URL."""
        if StreamingProtocol.WEB_RTC not in self.supported_protocols:
            raise ValueError("Device does not support WEB_RTC stream")
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.GenerateWebRtcStream",
            "params": {"offerSdp": offer_sdp},
        }
        response_data = await self._cmd.execute_json(data)
        results = response_data[RESULTS]
        return WebRtcStream(results, self._cmd)


class EventImage:
    """Provides access to an image in response to an event.

    Use a ?width or ?height query parameters to customize the resolution
    of the downloaded image. Only one of these parameters need to specified.
    The other parameter is scaled automatically according to the camera's
    aspect ratio.

    The token should be added as an HTTP header:
    Authorization: Basic <token>
    """

    def __init__(
        self,
        data: Mapping[str, Any],
        cmd: Command,
        event_image_type: EventImageContentType,
    ):
        """Initialize the EventImage."""
        self._data = data
        self._cmd = cmd
        self._event_image_type = event_image_type

    @property
    def event_image_type(self) -> EventImageContentType:
        """Return the type of event image."""
        return self._event_image_type

    @property
    def url(self) -> Optional[str]:
        """URL to download the camera image from."""
        return cast_optional(str, self._data.get(URL))

    @property
    def token(self) -> Optional[str]:
        """Token to use in the HTTP Authorization header when downloading."""
        return cast_optional(str, self._data.get(TOKEN))

    async def contents(
        self, width: Optional[int] = None, height: Optional[int] = None
    ) -> bytes:
        """Download the image bytes."""
        if width:
            fetch_url = f"{self.url}?width={width}"
        elif height:
            fetch_url = f"{self.url}?width={height}"
        else:
            assert self.url
            fetch_url = self.url
        return await self._cmd.fetch_image(fetch_url, basic_auth=self.token)


class EventImageCreator(ABC):
    """Responsible for turning events into an EventImage."""

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide an object that can be used to fetch media for an event."""


@TRAIT_MAP.register()
class CameraEventImageTrait(EventImageCreator):
    """This trait belongs to any device that generates images from events."""

    NAME = "sdm.devices.traits.CameraEventImage"

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraEventImageTrait."""
        self._data = data
        self._cmd = cmd

    async def generate_image(self, event_id: str) -> EventImage:
        """Provide a URL to download a camera image."""
        data = {
            "command": "sdm.devices.commands.CameraEventImage.GenerateImage",
            "params": {
                "eventId": event_id,
            },
        }
        response_data = await self._cmd.execute_json(data)
        results = response_data[RESULTS]
        return EventImage(results, self._cmd, EventImageType.IMAGE)

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from an event."""
        return await self.generate_image(event.event_id)


class EventImageGenerator(EventTrait, EventImageCreator, ABC):
    """Parenet class for a trait that generates an images from events."""

    @property
    @abstractmethod
    def event_type(self) -> str:
        """Event types supported by this trait."""

    async def generate_active_event_image(self) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        event = self.active_event
        if not event:
            _LOGGER.debug("No active event")
            return None
        return await self.generate_event_image(event)


@TRAIT_MAP.register()
class CameraMotionTrait(EventImageGenerator):
    """For any device that supports motion detection events."""

    NAME = "sdm.devices.traits.CameraMotion"
    EVENT_NAME = CameraMotionEvent.NAME
    event_type = EVENT_NAME
    event_image_creator: EventImageCreator | None = None

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraClipPreviewTrait."""
        super().__init__()
        self._event_image = CameraEventImageTrait({}, cmd)

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        if not isinstance(event, CameraMotionEvent):
            return None
        assert event.event_id
        if not self.event_image_creator:
            raise ValueError("Camera does not have trait to fetch snapshots")
        return await self.event_image_creator.generate_event_image(event)


@TRAIT_MAP.register()
class CameraPersonTrait(EventImageGenerator):
    """For any device that supports person detection events."""

    NAME = "sdm.devices.traits.CameraPerson"
    EVENT_NAME = CameraPersonEvent.NAME
    event_type = EVENT_NAME
    event_image_creator: EventImageCreator | None = None

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraClipPreviewTrait."""
        super().__init__()
        self._event_image = CameraEventImageTrait({}, cmd)

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        if not isinstance(event, CameraPersonEvent):
            return None
        assert event.event_id
        if not self.event_image_creator:
            raise ValueError("Camera does not have trait to fetch snapshots")
        return await self.event_image_creator.generate_event_image(event)


@TRAIT_MAP.register()
class CameraSoundTrait(EventImageGenerator):
    """For any device that supports sound detection events."""

    NAME = "sdm.devices.traits.CameraSound"
    EVENT_NAME = CameraSoundEvent.NAME
    event_type = EVENT_NAME
    event_image_creator: EventImageCreator | None = None

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraClipPreviewTrait."""
        super().__init__()

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        if not isinstance(event, CameraSoundEvent):
            return None
        assert event.event_id
        if not self.event_image_creator:
            raise ValueError("Camera does not have trait to fetch snapshots")
        return await self.event_image_creator.generate_event_image(event)


@TRAIT_MAP.register()
class CameraClipPreviewTrait(EventImageGenerator):
    """For any device that supports a clip preview."""

    NAME = "sdm.devices.traits.CameraClipPreview"
    EVENT_NAME = CameraClipPreviewEvent.NAME
    event_type = EVENT_NAME

    def __init__(self, data: Mapping[str, Any], cmd: Command):
        """Initialize CameraClipPreviewTrait."""
        super().__init__()
        self._cmd = cmd

    async def generate_event_image(self, event: ImageEventBase) -> Optional[EventImage]:
        """Provide a URL to download a camera image from the active event."""
        preview_event: CameraClipPreviewEvent | None = None
        if isinstance(event, CameraClipPreviewEvent):
            preview_event = event
        else:
            for session_event in event.session_events:
                if isinstance(session_event, CameraClipPreviewEvent):
                    preview_event = session_event
                    break
            if preview_event is None:
                _LOGGER.debug("Ignoring event, not the right type: %s", event)
                return None
        # Clip preview events have the url baked in without an additional
        # step to generate the image
        assert preview_event
        return EventImage(
            {"url": preview_event.preview_url}, self._cmd, EventImageType.CLIP_PREVIEW
        )



"""Base library for all traits."""

from __future__ import annotations

from typing import Any, Dict, Mapping, Optional

import aiohttp

from .auth import AbstractAuth
from .registry import Registry

DEVICE_TRAITS = "traits"

TRAIT_MAP = Registry()


class Command:
    """Base class for executing commands."""

    def __init__(self, device_id: str, auth: AbstractAuth):
        """Initialize Command."""
        self._device_id = device_id
        self._auth = auth

    async def execute(self, data: Mapping[str, Any]) -> aiohttp.ClientResponse:
        """Run the command."""
        assert self._auth
        return await self._auth.post(f"{self._device_id}:executeCommand", json=data)

    async def execute_json(self, data: Mapping[str, Any]) -> dict[str, Any]:
        """Run the command and return a json result."""
        assert self._auth
        return await self._auth.post_json(
            f"{self._device_id}:executeCommand", json=data
        )

    async def fetch_image(self, url: str, basic_auth: Optional[str] = None) -> bytes:
        """Fetch an image at the specified url."""
        headers: Dict[str, Any] = {}
        if basic_auth:
            headers = {"Authorization": f"Basic {basic_auth}"}
        resp = await self._auth.get(url, headers=headers)
        return await resp.read()


def _TraitsDict(
    traits: Mapping[str, Any], trait_map: Mapping[str, Any], cmd: Command
) -> Dict[str, Any]:
    d = {}
    for (trait, trait_data) in traits.items():
        if trait not in trait_map:
            continue
        cls = trait_map[trait]
        d[trait] = cls(trait_data, cmd)
    return d


def BuildTraits(
    traits: Mapping[str, Any], cmd: Command, device_type: Optional[str] = None
) -> Dict[str, Any]:
    """Build a trait map out of a response dict."""
    # There is a bug where doorbells do not return the DoorbellChime trait.  Simulate
    # that it was returned
    if device_type and device_type == "sdm.devices.types.DOORBELL":
        traits = dict(traits)
        traits["sdm.devices.traits.DoorbellChime"] = {}
    return _TraitsDict(traits, TRAIT_MAP, cmd)



#!/usr/bin/python3

"""Command line tool for the Google Nest Smart Device Management API.

You must configure your device as described:
  https://developers.google.com/nest/device-access/get-started

which will give you a project_id, client_id, and client_secret.  This tool
will do a one time setup to get an access token, then from there on will
cache the token on local disk.

Once authenticated, you can run commands like:

$ google_nest --project_id=<project_id> list
$ google_nest --project_id=<project_id> get <device_id>
"""

import argparse
import asyncio
import errno
import json
import logging
import os
import pickle
from typing import List, Optional, cast

import yaml
from aiohttp import ClientSession
from google.auth.credentials import Credentials
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow

from .auth import AbstractAuth
from .camera_traits import CameraLiveStreamTrait
from .device import Device
from .event import EventMessage
from .google_nest_api import GoogleNestAPI
from .google_nest_subscriber import (
    API_URL,
    OAUTH2_AUTHORIZE_FORMAT,
    OAUTH2_TOKEN,
    SDM_SCOPES,
    GoogleNestSubscriber,
)
from .structure import Structure
from .thermostat_traits import (
    ThermostatEcoTrait,
    ThermostatModeTrait,
    ThermostatTemperatureSetpointTrait,
)

# Define command line arguments
parser = argparse.ArgumentParser(
    description="Command line tool for Google Nest SDM API"
)
parser.add_argument("--project_id", required=True, help="Device Access program id")
parser.add_argument("--client_id", help="OAuth credentials client_id")
parser.add_argument("--client_secret", help="OAuth credentials client_secret")
parser.add_argument(
    "--token_cache",
    help="File storage for long lived creds",
    default="~/.config/google_nest/token_cache",
)
parser.add_argument(
    "-v", "--verbose", help="Increase output verbosity", action="store_true"
)
parser.add_argument(
    "--output_type",
    type=str,
    choices=["json", "yaml"],
    help="Change the output type from json or yaml (default).",
    default="yaml",
)

cmd_parser = parser.add_subparsers(dest="command", required=True)
list_structures_parser = cmd_parser.add_parser("list_structures")
list_devices_parser = cmd_parser.add_parser("list_devices")
get_structure_parser = cmd_parser.add_parser("get_structure")
get_structure_parser.add_argument("structure_id")
get_device_parser = cmd_parser.add_parser("get_device")
get_device_parser.add_argument("device_id")
set_mode_parser = cmd_parser.add_parser(
    "set_mode", description="Change the thermostat mode."
)
set_mode_parser.add_argument("device_id")
set_mode_parser.add_argument(
    "mode",
    help="The mode to change the thermostat to.",
    choices=["MANUAL_ECO", "HEAT", "COOL", "HEATCOOL", "OFF"],
)
set_heat_parser = cmd_parser.add_parser(
    "set_heat", description="Sets the target temperature when in HEAT mode."
)
set_heat_parser.add_argument("device_id")
set_heat_parser.add_argument("heat", type=float)
set_cool_parser = cmd_parser.add_parser(
    "set_cool", help="Sets the target temperature when in COOL mode."
)
set_cool_parser.add_argument("device_id")
set_cool_parser.add_argument(
    "cool",
    type=float,
    help="The target temperature to set when the thermostat is in COOL mode.",
)
set_range_parser = cmd_parser.add_parser(
    "set_range", help="Sets the min/max temperature when in HEATCOOL mode."
)
set_range_parser.add_argument("device_id")
set_range_parser.add_argument(
    "heat", type=float, help="The minimum target temperature to set."
)
set_range_parser.add_argument(
    "cool", type=float, help="The maximum target temperature to set."
)
generate_rtsp_stream_parser = cmd_parser.add_parser("generate_rtsp_stream")
generate_rtsp_stream_parser.add_argument("device_id")
generate_web_rtc_stream_parser = cmd_parser.add_parser("generate_web_rtc_stream")
generate_web_rtc_stream_parser.add_argument("device_id")
generate_web_rtc_stream_parser.add_argument("offer_file")
subscribe_parser = cmd_parser.add_parser("subscribe")
subscribe_parser.add_argument("subscription_id")
subscribe_parser.add_argument("device_id", nargs="?")


class Auth(AbstractAuth):
    """Implementation of AbstractAuth that uses the token cache."""

    def __init__(
        self,
        websession: ClientSession,
        user_creds: Credentials,
        api_url: str,
    ):
        """Initialize Google Nest Device Access auth."""
        super().__init__(websession, api_url)
        self._user_creds = user_creds

    async def async_get_access_token(self) -> str:
        """Return a valid access token."""
        return cast(str, self._user_creds.token)

    async def async_get_creds(self) -> Credentials:
        """Return valid OAuth creds."""
        return self._user_creds


def CreateCreds(args: argparse.Namespace) -> Credentials:
    """Run an interactive flow to get OAuth creds."""
    creds = None
    token_cache = os.path.expanduser(args.token_cache)
    if os.path.exists(token_cache):
        with open(token_cache, "rb") as token:
            creds = pickle.load(token)

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            if not args.client_id or not args.client_secret:
                raise ValueError("Required flag --client_id or --client_secret missing")
            client_config = {
                "installed": {
                    "client_id": args.client_id,
                    "client_secret": args.client_secret,
                    "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob"],
                    "auth_uri": OAUTH2_AUTHORIZE_FORMAT.format(
                        project_id=args.project_id
                    ),
                    "token_uri": OAUTH2_TOKEN,
                },
            }
            app_flow = InstalledAppFlow.from_client_config(
                client_config, scopes=SDM_SCOPES
            )
            creds = app_flow.run_console()
        # Save the credentials for the next run
        if not os.path.exists(os.path.dirname(token_cache)):
            try:
                os.makedirs(os.path.dirname(token_cache))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        with open(token_cache, "wb") as token:
            pickle.dump(creds, token)
    return creds


def PrintStructure(structure: Structure, output_type: str) -> None:
    """Print the structure."""
    if output_type == "json":
        print(json.dumps(structure.raw_data))
    else:
        print(yaml.dump(structure.raw_data))


def PrintDevice(device: Device, output_type: str) -> None:
    """Print the device."""
    if output_type == "json":
        print(json.dumps(device.raw_data))
    else:
        print(yaml.dump(device.raw_data))


class SubscribeCallback:
    """Print the event message."""

    def __init__(self, output_type: Optional[str] = None) -> None:
        """Initialize SubscribeCallback."""
        self._output_type = output_type

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Handle an EventMessage."""
        if self._output_type == "json":
            print(json.dumps(event_message.raw_data))
        else:
            print(yaml.dump(event_message.raw_data))


class DeviceWatcherCallback:
    """Print the event message."""

    def __init__(self, device: Device, output_type: str) -> None:
        """Initialize DeviceWatcherCallback."""
        self._device = device
        self._output_type = output_type

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Handle an EventMessage."""
        print(f"event_id: {event_message.event_id}")
        print("Current device state:")
        PrintDevice(self._device, self._output_type)
        print("")


async def RunTool(args: argparse.Namespace, user_creds: Credentials) -> None:
    """Run the command."""
    async with ClientSession() as client:
        auth = Auth(client, user_creds, API_URL)
        api = GoogleNestAPI(auth, args.project_id)

        if args.command == "list_structures":
            structures: List[Structure] = await api.async_get_structures()
            for s in structures:
                PrintStructure(s, args.output_type)
            return

        if args.command == "get_structure":
            structure: Optional[Structure] = await api.async_get_structure(
                args.structure_id
            )
            assert structure
            PrintStructure(structure, args.output_type)
            return

        if args.command == "list_devices":
            devices = await api.async_get_devices()
            for d in devices:
                PrintDevice(d, args.output_type)
            return

        if args.command == "subscribe":
            logging.info("Subscription: %s", args.subscription_id)
            subscriber = GoogleNestSubscriber(
                auth, args.project_id, args.subscription_id
            )
            if args.device_id:
                device_manager = await subscriber.async_get_device_manager()
                dev = device_manager.devices[args.device_id]
                dev_callback = DeviceWatcherCallback(dev, args.output_type)
                dev.add_event_callback(dev_callback.async_handle_event)
            else:
                sub_callback = SubscribeCallback(args.output_type)
                subscriber.set_update_callback(sub_callback.async_handle_event)
            await subscriber.start_async()
            try:
                while True:
                    await asyncio.sleep(10)
            except KeyboardInterrupt:
                subscriber.stop_async()

        # All other commands require a device_id
        device: Optional[Device] = await api.async_get_device(args.device_id)
        assert device

        if args.command == "get_device":
            PrintDevice(device, args.output_type)

        if args.command == "set_mode":
            mode = args.mode
            trait = device.traits[ThermostatModeTrait.NAME]
            if mode == "MANUAL_ECO":
                trait = device.traits[ThermostatEcoTrait.NAME]
            resp = await trait.set_mode(mode)
            print(await resp.text())

        if args.command == "set_heat":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_heat(args.heat)
            print(await resp.text())

        if args.command == "set_cool":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_cool(args.cool)
            print(await resp.text())

        if args.command == "set_range":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_range(args.heat, args.cool)
            print(await resp.text())

        if args.command == "generate_rtsp_stream":
            trait = device.traits[CameraLiveStreamTrait.NAME]
            stream = await trait.generate_rtsp_stream()
            print(f"URL: {stream.rtsp_stream_url}")
            print(f"Stream Token: {stream.stream_token}")
            print(f"Expires At: {stream.expires_at}")

        if args.command == "generate_web_rtc_stream":
            trait = device.traits[CameraLiveStreamTrait.NAME]
            offer_sdp = None
            if args.offer_file:
                f = open(args.offer_file, "r")
                offer_sdp = f.read()
            stream = await trait.generate_web_rtc_stream(offer_sdp)
            print(f"Answer SDP: {stream.answer_sdp}")
            print(f"Media Session Id: {stream.media_session_id}")
            print(f"Expires At: {stream.expires_at}")


def main() -> None:
    """Nest command line tool."""
    args: argparse.Namespace = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    user_creds = CreateCreds(args)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(RunTool(args, user_creds))
    loop.close()


if __name__ == "__main__":
    main()



"""A device from the Smart Device Management API."""

from __future__ import annotations

import datetime
import logging
from typing import Any, Awaitable, Callable, Dict, List, Mapping, Optional, cast

# Import traits for registration
from . import camera_traits  # noqa: F401
from . import device_traits  # noqa: F401
from . import doorbell_traits  # noqa: F401
from . import thermostat_traits  # noqa: F401
from .auth import AbstractAuth
from .event import EventMessage, EventProcessingError, EventTrait
from .event_media import EventMediaManager
from .traits import BuildTraits, Command
from .typing import cast_assert

_LOGGER = logging.getLogger(__name__)

DEVICE_NAME = "name"
DEVICE_TYPE = "type"
DEVICE_TRAITS = "traits"
DEVICE_PARENT_RELATIONS = "parentRelations"
PARENT = "parent"
DISPLAYNAME = "displayName"


def _MakeEventTraitMap(
    traits: Mapping[str, Any]
) -> Dict[str, camera_traits.EventImageGenerator]:
    event_trait_map: Dict[str, Any] = {}
    for (trait_name, trait) in traits.items():
        if not isinstance(trait, camera_traits.EventImageGenerator):
            continue
        event_trait_map[trait.event_type] = trait
    return event_trait_map


class Device:
    """Class that represents a device object in the Google Nest SDM API."""

    def __init__(
        self,
        raw_data: Mapping[str, Any],
        traits: Dict[str, Any],
        event_media_manager: EventMediaManager,
    ) -> None:
        """Initialize a device."""
        self._raw_data = raw_data
        self._traits = traits
        self._trait_event_ts: Dict[str, datetime.datetime] = {}
        self._relations = {}
        for relation in self._raw_data.get(DEVICE_PARENT_RELATIONS, []):
            if PARENT not in relation or DISPLAYNAME not in relation:
                continue
            self._relations[relation[PARENT]] = relation[DISPLAYNAME]
        self._callbacks: List[Callable[[EventMessage], Awaitable[None]]] = []
        self._event_media_manager = event_media_manager

    @staticmethod
    def MakeDevice(raw_data: Mapping[str, Any], auth: AbstractAuth) -> Device:
        """Create a device with the appropriate traits."""
        device_id = raw_data.get(DEVICE_NAME)
        if not device_id:
            raise ValueError(f"raw_data missing field '{DEVICE_NAME}'")
        cmd = Command(device_id, auth)
        traits = raw_data.get(DEVICE_TRAITS, {})
        traits_dict = BuildTraits(traits, cmd, raw_data.get(DEVICE_TYPE))

        # Hack to wire up camera traits to the event image generator
        event_image_trait: camera_traits.EventImageCreator | None = None
        if camera_traits.CameraEventImageTrait.NAME in traits_dict:
            event_image_trait = traits_dict[camera_traits.CameraEventImageTrait.NAME]
        elif camera_traits.CameraClipPreviewTrait.NAME in traits_dict:
            event_image_trait = traits_dict[camera_traits.CameraClipPreviewTrait.NAME]
        if event_image_trait:
            for trait_class in traits_dict.values():
                if hasattr(trait_class, "event_image_creator"):
                    trait_class.event_image_creator = event_image_trait
        event_trait_map = _MakeEventTraitMap(traits_dict)
        event_media_manager = EventMediaManager(
            device_id,
            traits_dict,
            event_trait_map,
            support_fetch=(event_image_trait is not None),
        )
        return Device(raw_data, traits_dict, event_media_manager)

    @property
    def name(self) -> str:
        """Resource name of the device such as 'enterprises/XYZ/devices/123'."""
        return cast_assert(str, self._raw_data[DEVICE_NAME])

    @property
    def type(self) -> str:
        """Type of device for display purposes.

        The device type should not be used to deduce or infer functionality of
        the actual device it is assigned to. Instead, use the returned traits for
        the device.
        """
        return cast_assert(str, self._raw_data[DEVICE_TYPE])

    @property
    def traits(self) -> Dict[str, Any]:
        """Return a trait mixin or None."""
        return self._traits

    def _traits_data(self, trait: str) -> Dict[str, Any]:
        """Return the raw dictionary for the specified trait."""
        traits_dict = self._raw_data.get(DEVICE_TRAITS, {})
        return cast(Dict[str, Any], traits_dict.get(trait, {}))

    @property
    def parent_relations(self) -> dict:
        """Room or structure for the device."""
        return self._relations

    def add_update_listener(self, target: Callable[[], None]) -> Callable[[], None]:
        """Register a simple event listener notified on updates.

        The return value is a callable that will unregister the callback.
        """

        async def handle_event(event_message: EventMessage) -> None:
            target()

        return self.add_event_callback(handle_event)

    def add_event_callback(
        self, target: Callable[[EventMessage], Awaitable[None]]
    ) -> Callable[[], None]:
        """Register an event callback for updates to this device.

        The return value is a callable that will unregister the callback.
        """
        self._callbacks.append(target)

        def remove_callback() -> None:
            """Remove the event_callback."""
            self._callbacks.remove(target)

        return remove_callback

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Process an event from the pubsub subscriber."""
        _LOGGER.debug(
            "Processing update %s @ %s", event_message.event_id, event_message.timestamp
        )
        if not event_message.resource_update_name:
            raise EventProcessingError("Event was not resource update event")
        if self.name != event_message.resource_update_name:
            raise EventProcessingError(
                f"Mismatch {self.name} != {event_message.resource_update_name}"
            )
        self._async_handle_traits(event_message)
        await self._event_media_manager.async_handle_events(event_message)
        for callback in self._callbacks:
            await callback(event_message)

    def _async_handle_traits(self, event_message: EventMessage) -> None:
        traits = event_message.resource_update_traits
        if not traits:
            return
        _LOGGER.debug("Trait update %s", traits)
        for (trait_name, trait) in traits.items():
            # Discard updates older than prior events
            # Note: There is still a race where traits read from the API on
            # startup are overwritten with old messages. We assume that the
            # event messages will eventually correct that.
            if trait_name in self._trait_event_ts:
                ts = self._trait_event_ts[trait_name]
                if ts > event_message.timestamp:
                    continue
            self._traits[trait_name] = trait
            self._trait_event_ts[trait_name] = event_message.timestamp

    @property
    def event_media_manager(self) -> EventMediaManager:
        return self._event_media_manager

    @property
    def active_event_trait(self) -> Optional[EventTrait]:
        """Return trait with the most recently received active event."""
        return self._event_media_manager.active_event_trait

    @property
    def raw_data(self) -> Dict[str, Any]:
        """Return raw data for the device."""
        return dict(self._raw_data)



"""Libraries for helping with typing API responses."""

from typing import Any, Optional, Type, TypeVar

T = TypeVar("T")


def cast_assert(t: Type[T], data: Any) -> T:
    """Function to aid in extracting type values from API responses."""
    assert isinstance(data, t), f"Expected data with type {t} but was {type(data)}"
    return data


def cast_optional(t: Type[T], data: Any) -> Optional[T]:
    """Function to aid in extracting type values from API responses."""
    if data is None:
        return None
    assert isinstance(data, t), f"Expected data with type {t} but was {type(data)}"
    return data
